/*
 * packet_headers.c  -dlevette
 *
 * packet header construction routines
*/
#include <stdio.h>
#include <string.h>        // memcpy()
#include <stdlib.h>

#include <net/ethernet.h>  // struct ether_header
#include <arpa/inet.h>     // htons() htonl()

#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <linux/icmp.h>
#include <netinet/in.h>
#include <net/if_arp.h>

#include "raw_stuff.h"


struct ether_header *build_ether_hdr(char *src_mac, char *dest_mac, int ether_proto)
{
	struct ether_header *ether_head;
	/* allocate memory */
	ether_head = malloc(sizeof(struct ether_header));

	/* Copy source and destination mac addresses into struct feilds */
	memcpy(ether_head->ether_shost, (void *)ether_aton(src_mac), ETH_ALEN);
	memcpy(ether_head->ether_dhost, (void *)ether_aton(dest_mac), ETH_ALEN);

	/* Copy over ETHERTYPE */
	ether_head->ether_type = htons(ether_proto);

	return (ether_head);
}

struct ip *build_ip_hdr(char *src_addr, char *dest_addr, u_short ttl, 
		char proto, int frag_off, int trans_sz, int payload_sz, u_short ip_id)
{
	struct ip *ip_hdr;
	
	ip_hdr = malloc(sizeof(struct ip));
	memset(ip_hdr, 0, sizeof(struct ip));
	
	ip_hdr->ip_hl   = 5;
	ip_hdr->ip_v    = 4; 
	ip_hdr->ip_tos  = 0; 
	ip_hdr->ip_len  = sizeof(struct ip) + trans_sz + payload_sz;
	ip_hdr->ip_id   = htons(ip_id); 
	ip_hdr->ip_off  = htons(frag_off);
	ip_hdr->ip_ttl  = ttl;
	ip_hdr->ip_p    = proto;
	ip_hdr->ip_sum  = 0;
	ip_hdr->ip_src.s_addr = inet_addr(src_addr);
	ip_hdr->ip_dst.s_addr = inet_addr(dest_addr);
	
	ip_hdr->ip_sum = calc_csum((u_char *)ip_hdr, sizeof(struct ip));
	
	return(ip_hdr);
}

struct tcphdr *build_tcp_hdr(u_short src_port, u_short dest_port, u_short win_sz,
			     int TCP_FIN, int TCP_SYN, int TCP_RST, int TCP_PSH,
			     int TCP_ACK, int TCP_URG, u_short urg, struct ip *ip_hdr,
				  	char *payload, int payload_sz)
{
	struct pseudo *hdr;
	int size = 0;
	u_char *psu_tcp;
	
	struct tcphdr *tcp_hdr;
	tcp_hdr = malloc(sizeof(struct tcphdr));
	
	tcp_hdr->source  = htons(src_port);
	tcp_hdr->dest    = htons(dest_port);
	tcp_hdr->seq     = htonl(88); // FIXME: now mwah ha ha
	tcp_hdr->ack_seq = htonl(77); // FIXME:
	tcp_hdr->res1    = 0;
	tcp_hdr->doff    = 5;
	
	/* Flags */
	tcp_hdr->fin     = TCP_FIN;
	tcp_hdr->syn	 = TCP_SYN;
	tcp_hdr->rst	 = TCP_RST;
	tcp_hdr->psh	 = TCP_PSH;
	tcp_hdr->ack 	 = TCP_ACK;
	tcp_hdr->urg	 = TCP_URG;
	tcp_hdr->window  = htons(payload_sz);
	tcp_hdr->check	 = 0;   
	tcp_hdr->urg_ptr = urg;

	/* Calculate the tcp checksum, this is done over the entire tcp hdr + */
        /* the data + a psuedo header	*/
	hdr = build_pseudo(ip_hdr);
	
	/* Allocate mem for header + pseudo + data */
	size = payload_sz + sizeof(struct tcphdr) + sizeof(struct pseudo);
	psu_tcp = malloc(size);
	memset(psu_tcp, 0, size);
	
	/* Fill in length of header + data */
	hdr->len = htons(size - sizeof(struct pseudo));
	
	/* Build data to be checksumed */	
	memcpy(psu_tcp, hdr, sizeof(struct pseudo));
	memcpy(psu_tcp + sizeof(struct pseudo), tcp_hdr, sizeof(struct tcphdr));
	memcpy(psu_tcp + sizeof(struct pseudo) + sizeof(struct tcphdr), payload,
		payload_sz);
	
	tcp_hdr->check = calc_csum((u_char *)psu_tcp, size);
	
	free(hdr);
	free(psu_tcp);
	
	return(tcp_hdr);
}

struct udphdr *build_udp_hdr(u_short src_port, u_short dest_port, 
				int trans_sz, int payload_sz, struct ip *ip_hdr, 
					char *payload)
{
	struct pseudo *hdr;
	int size = 0;
	u_char *psu_udp;

	struct udphdr *udp_hdr;
	udp_hdr = malloc(sizeof(struct udphdr));

	udp_hdr->source = htons(src_port);
	udp_hdr->dest   = htons(dest_port);
	udp_hdr->len    = htons(trans_sz + payload_sz);
	udp_hdr->check  = 0;

	/* Calculate the udp checksum, done in the same way as tcp */
	hdr = build_pseudo(ip_hdr);
	
	size = payload_sz + sizeof(struct udphdr) + sizeof(struct pseudo);
	psu_udp = malloc(size);
	memset(psu_udp, 0, size);
	
	hdr->len = htons(size - sizeof(struct pseudo));
	
	memcpy(psu_udp, hdr, sizeof(struct pseudo));
	memcpy(psu_udp + sizeof(struct pseudo), udp_hdr, sizeof(struct udphdr));
	memcpy(psu_udp + sizeof(struct pseudo) + sizeof(struct udphdr), payload,
		payload_sz);
	
	udp_hdr->check = calc_csum((u_char *)psu_udp, size);
	
	free(hdr);
	free(psu_udp);
		
	return(udp_hdr);
}

struct icmphdr *build_icmp_hdr(u_short icmp_type, u_short icmp_code,
				u_short icmp_id, u_short icmp_seq, struct ip *ip_hdr, 
					char *payload, int payload_sz)
{	
	int size = 0;
	u_char *psu_icmp;
	
	struct icmphdr *icmp_hdr;
	icmp_hdr = malloc(sizeof(struct icmphdr));
	
	icmp_hdr->type 	   = icmp_type;
	icmp_hdr->code     = icmp_code;
	icmp_hdr->checksum = 0;
	
	/* Calculate the icmp checksum */
	size = payload_sz + sizeof(struct icmphdr);
	psu_icmp = malloc(size);
	memset(psu_icmp, 0, size);
	
	memcpy(psu_icmp, icmp_hdr, sizeof(struct icmphdr));
	memcpy(psu_icmp + sizeof(struct icmphdr), payload, payload_sz);
	
	icmp_hdr->checksum = calc_csum((u_char *)psu_icmp, size);
	
	free(psu_icmp);

	return(icmp_hdr);
}

struct arphdr *build_arp_hdr(u_short arp_optc)
{
	struct arphdr *arp_hdr;
	arp_hdr = malloc(sizeof(struct arphdr));
		
	arp_hdr->ar_hrd = htons(ARPHRD_ETHER);
	arp_hdr->ar_pro = htons(ETHERTYPE_IP);
	arp_hdr->ar_hln = ETH_ALEN;
	arp_hdr->ar_pln = 4;
	arp_hdr->ar_op  = htons(arp_optc);
	
	return(arp_hdr);	
}

struct arp_ethip_hdr *build_arp_ethip(char *send_mac, char *send_ip, char *targ_mac,
					char *targ_ip)
{
	struct arp_ethip_hdr *arp_ip_hdr;
	arp_ip_hdr = malloc(sizeof(struct arp_ethip_hdr));
	
	memcpy(arp_ip_hdr->sender_mac, (void *)ether_aton(send_mac), ETH_ALEN);
	arp_ip_hdr->sender_ip.s_addr = inet_addr(send_ip);
	memcpy(arp_ip_hdr->target_mac, (void *)ether_aton(targ_mac), ETH_ALEN);
	arp_ip_hdr->target_ip.s_addr = inet_addr(targ_ip);
	
	return(arp_ip_hdr);
}

struct pseudo *build_pseudo(struct ip *ip_hdr)
{

	struct pseudo *hdr;
	hdr = malloc(sizeof(struct pseudo));
	
	hdr->src_ip  = ip_hdr->ip_src;
	hdr->dest_ip = ip_hdr->ip_dst;
	hdr->reserved = 0;
	hdr->protocol = ip_hdr->ip_p;
	hdr->len = 0;
	
	return(hdr);
}
